/*
 * GPIO driver for Ingenic X1600.
 * (See below for additional copyright and licensing notices.)
 *
 * Copyright (C) 2017, 2023 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 *
 *
 * Subject to other copyrights, being derived from the bcm2835.cc and
 * omap.cc GPIO driver implementations.
 *
 * This file is part of TUD:OS and distributed under the terms of the
 * GNU General Public License 2.
 * Please see the COPYING-GPL-2 file for details.
 */

#include <l4/sys/icu.h>
#include <l4/util/util.h>
#include <l4/devices/hw_mmio_register_block.h>

#include "gpio-x1600.h"

/*
GPIO register offsets (x in A..D).

Register summary:

PxINT         0 (function/GPIO)                                             1 (interrupt)
PxMSK         0 (function)                        1 (GPIO)                  0 (IRQ enable)/1 (IRQ disable)
PxPAT1        0 (function 0/1)  1 (function 2/3)  0 (output)    1 (input)   0 (level trigger) 1 (edge trigger)
PxPAT0        0 (function 0)    0 (function 2)    0 (output value 0)        0 (low level)     0 (falling edge)
              1 (function 1)    1 (function 3)    1 (output value 1)        1 (high level)    1 (rising edge)
*/

enum Regs
{
  Pin_level           = 0x000,  // PxPINL (read-only)

  Port_int            = 0x010,  // PxINT
  Port_int_set        = 0x014,  // PxINTS
  Port_int_clear      = 0x018,  // PxINTC

  Irq_mask            = 0x020,  // PxMSK (for PxINT == 1)
  Irq_mask_set        = 0x024,  // PxMSKS
  Irq_mask_clear      = 0x028,  // PxMSKC
  Port_gpio           = 0x020,  // PxMSK (for PxINT == 0)
  Port_gpio_set       = 0x024,  // PxMSKS
  Port_gpio_clear     = 0x028,  // PxMSKC

  Port_trigger        = 0x030,  // PxPAT1 (for PxINT == 1)
  Port_trigger_set    = 0x034,  // PxPAT1S
  Port_trigger_clear  = 0x038,  // PxPAT1C
  Port_dir            = 0x030,  // PxPAT1 (for PxINT == 0, PxMSK == 1)
  Port_dir_set        = 0x034,  // PxPAT1S
  Port_dir_clear      = 0x038,  // PxPAT1C
  Port_group1         = 0x030,  // PxPAT1 (for PxINT == 0, PxMSK == 0)
  Port_group1_set     = 0x034,  // PxPAT1S
  Port_group1_clear   = 0x038,  // PxPAT1C

  Port_level          = 0x040,  // PxPAT0 (for PxINT == 1)
  Port_level_set      = 0x044,  // PxPAT0S
  Port_level_clear    = 0x048,  // PxPAT0C
  Port_data           = 0x040,  // PxPAT0 (for PxINT == 0, PxMSK == 1, PxPAT1 == 0)
  Port_data_set       = 0x044,  // PxPAT0S
  Port_data_clear     = 0x048,  // PxPAT0C
  Port_group0         = 0x040,  // PxPAT0 (for PxINT == 0, PxMSK == 0)
  Port_group0_set     = 0x044,  // PxPAT0S
  Port_group0_clear   = 0x048,  // PxPAT0C

  Irq_flag            = 0x050,  // PxFLG (read-only)
  Irq_flag_clear      = 0x058,  // PxFLGC

  // Only the following registers differ from the JZ4780. The dual-edge
  // registers being added to the X1600, with the pull-up/down registers being
  // relocated and their sense changed from disable to enable.

  Pull_edge           = 0x070,  // PxEDG
  Pull_edge_set       = 0x074,  // PxEDGS
  Pull_edge_clear     = 0x078,  // PxEDGC

  Pull_enable         = 0x080,  // PxPE
  Pull_enable_set     = 0x084,  // PxPES
  Pull_enable_clear   = 0x088,  // PxPEC

  // The shadow port Z is available at offset 0x700 and supports the INTS, INTC,
  // MSKS, MSKC, PAT1S, PAT1C, PAT0S, PAT0C registers, along with the following.

  Shadow_transfer     = 0x0f0,  // PzGID2LD
};



// IRQ control for each GPIO pin.

Gpio_x1600_irq_pin::Gpio_x1600_irq_pin(unsigned pin, Hw::Register_block<32> const &regs,
                                       Hw::Register_block<32> const &shadow_regs,
                                       uint8_t port_number)
: _pin(pin), _regs(regs), _shadow_regs(shadow_regs), _port_number(port_number),
  _shadow(true)
{}

Gpio_x1600_irq_pin::Gpio_x1600_irq_pin(unsigned pin, Hw::Register_block<32> const &regs)
: _pin(pin), _regs(regs), _shadow(false)
{}

void
Gpio_x1600_irq_pin::write_reg_pin(unsigned reg)
{
  // Write the pin bit to the register, setting or clearing the pin
  // depending on the register chosen.

  if (_shadow)
    _shadow_regs[reg] = _pin_bit(_pin);
  else
    _regs[reg] = _pin_bit(_pin);
}

void Gpio_x1600_irq_pin::do_mask()
{
  // Set the interrupt bit in the PxIM register.

  write_reg_pin(Irq_mask_set);
}

void Gpio_x1600_irq_pin::do_unmask()
{
  // Clear the interrupt bit in the PxIM register, first also clearing the
  // flag bit in the PxFLG register to allow interrupts to be delivered.

  write_reg_pin(Irq_flag_clear);
  write_reg_pin(Irq_mask_clear);
}

bool Gpio_x1600_irq_pin::do_set_mode(unsigned mode)
{
  // Standard comment found for this method:
  // this operation touches multiple mmio registers and is thus
  // not atomic, that's why we first mask the IRQ and if it was
  // enabled we unmask it after we have changed the mode

  // The X1600 provides a special port Z that allows changes to be made and then
  // committed atomically using PzGID2LD.

  if (!_shadow && enabled())
    do_mask();

  // Do the PxINT, PxPAT1 and PxPAT0 configuration.

  switch(mode)
  {
    case L4_IRQ_F_LEVEL_HIGH:
      write_reg_pin(Port_int_set);
      write_reg_pin(Port_trigger_clear);
      write_reg_pin(Port_level_set);
      break;
    case L4_IRQ_F_LEVEL_LOW:
      write_reg_pin(Port_int_set);
      write_reg_pin(Port_trigger_clear);
      write_reg_pin(Port_level_clear);
      break;
    case L4_IRQ_F_POS_EDGE:
      write_reg_pin(Port_int_set);
      write_reg_pin(Port_trigger_set);
      write_reg_pin(Port_level_set);
      break;
    case L4_IRQ_F_NEG_EDGE:
      write_reg_pin(Port_int_set);
      write_reg_pin(Port_trigger_set);
      write_reg_pin(Port_level_clear);
      break;

    default:
      return false;
  }

  if (_shadow)
    _shadow_regs[Shadow_transfer] = _port_number;
  else if (enabled())
    do_unmask();

  return true;
}

int Gpio_x1600_irq_pin::clear()
{
  // Obtain the flag status for the pin, clearing it if set.

  l4_uint32_t e = _regs[Irq_flag] & (1UL << _pin);
  if (e)
    _regs[Irq_flag_clear] = e;

  return (e >> _pin);
}

bool Gpio_x1600_irq_pin::enabled()
{
  return true;
}



// Initialise the GPIO controller.

Gpio_x1600_chip::Gpio_x1600_chip(l4_addr_t start, l4_addr_t end,
                                 unsigned nr_pins,
                                 l4_uint32_t pull_ups, l4_uint32_t pull_downs,
                                 l4_addr_t shadow_start, l4_addr_t shadow_end,
                                 uint8_t port_number)
: _start(start), _end(end),
  _nr_pins(nr_pins),
  _pull_ups(pull_ups), _pull_downs(pull_downs),
  _shadow_start(shadow_start), _shadow_end(shadow_end),
  _port_number(port_number)
{
  _regs = new Hw::Mmio_register_block<32>(_start);

  if (_shadow_start)
  {
    _shadow_regs = new Hw::Mmio_register_block<32>(_shadow_start);
    _shadow = true;
  }
  else
    _shadow = false;
}

void
Gpio_x1600_chip::write_reg_pin(unsigned reg, unsigned pin)
{
  // Write the pin bit to the register, setting or clearing the pin
  // depending on the register chosen.

  if (_shadow)
    _shadow_regs[reg] = _pin_bit(pin);
  else
    _regs[reg] = _pin_bit(pin);
}

// Return the value of a pin.

int
Gpio_x1600_chip::get(unsigned pin)
{
  if (pin >= _nr_pins)
    throw -L4_EINVAL;

  l4_uint32_t val = _regs[Pin_level];
  return (val >> _pin_shift(pin)) & 1;
}

// Return multiple pin values.

unsigned
Gpio_x1600_chip::multi_get(unsigned offset)
{
  _reg_offset_check(offset);
  return _regs[Pin_level];
}

// Set the value of a pin.

void
Gpio_x1600_chip::set(unsigned pin, int value)
{
  if (pin >= _nr_pins)
    throw -L4_EINVAL;

  l4_uint32_t reg_set = value ? Port_data_set : Port_data_clear;
  _regs[reg_set] = _pin_bit(pin);
}

// Set multiple pin values.

void
Gpio_x1600_chip::multi_set(Pin_slice const &mask, unsigned data)
{
  _reg_offset_check(mask.offset);
  if (mask.mask & data)
    _regs[Port_data_set] = (mask.mask & data);
  if (mask.mask & ~data)
    _regs[Port_data_clear] = (mask.mask & ~data);
}

// Set a pin up with the given mode and value (if appropriate).

void
Gpio_x1600_chip::setup(unsigned pin, unsigned mode, int value)
{
  if (pin >= _nr_pins)
    throw -L4_EINVAL;

  config(pin, mode);

  if (mode == Output)
    set(pin, value);
}

// Configuration of a pin using the generic input/output/IRQ mode.

void
Gpio_x1600_chip::config(unsigned pin, unsigned mode)
{
  switch (mode)
  {
    case Input:
      write_reg_pin(Port_int_clear, pin);
      write_reg_pin(Port_gpio_set, pin);
      write_reg_pin(Port_dir_set, pin);
      break;
    case Output:
      write_reg_pin(Port_int_clear, pin);
      write_reg_pin(Port_gpio_set, pin);
      write_reg_pin(Port_dir_clear, pin);
      break;
    case Irq:
      write_reg_pin(Port_int_set, pin);
      // Other details depend on the actual trigger mode.
      break;
    default:
      break;
  }

  if (_shadow)
    _shadow_regs[Shadow_transfer] = _port_number;
}

// Pull-up/down configuration for a pin.

void
Gpio_x1600_chip::config_pull(unsigned pin, unsigned mode)
{
  if (pin >= _nr_pins)
    throw -L4_EINVAL;

  switch (mode)
  {
    case Pull_none:
      _regs[Pull_enable_clear] = _pin_bit(pin);
      break;
    case Pull_down:
      if (_pin_bit(pin) & _pull_downs)
        _regs[Pull_enable_set] = _pin_bit(pin);
      break;
    case Pull_up:
      if (_pin_bit(pin) & _pull_ups)
        _regs[Pull_enable_set] = _pin_bit(pin);
      break;
    default:
      // Invalid pull-up/down mode for pin.
      throw -L4_EINVAL;
  }
}

// Pin function configuration.

void
Gpio_x1600_chip::config_pad(unsigned pin, unsigned func, unsigned value)
{
  if (pin >= _nr_pins)
    throw -L4_EINVAL;

  if (value > 3)
    throw -L4_EINVAL;

  switch (func)
  {
    // Support two different outputs.

    case Hw::Gpio_chip::Function_gpio:
      write_reg_pin(Port_int_clear, pin);
      write_reg_pin(Port_gpio_set, pin);
      write_reg_pin(value & 1 ? Port_data_set : Port_data_clear, pin);
      break;

    // Support four different device functions.

    case Hw::Gpio_chip::Function_alt:
      write_reg_pin(Port_int_clear, pin);
      write_reg_pin(Port_gpio_clear, pin);
      write_reg_pin(value & 2 ? Port_group1_set : Port_group1_clear, pin);
      write_reg_pin(value & 1 ? Port_group0_set : Port_group0_clear, pin);
      break;
    default:
      throw -L4_EINVAL;
  }

  if (_shadow)
    _shadow_regs[Shadow_transfer] = _port_number;
}

// Obtain a pin's configuration from a register in the supplied value.

void
Gpio_x1600_chip::config_get(unsigned pin, unsigned reg, unsigned *value)
{
  if (pin >= _nr_pins)
    throw -L4_EINVAL;

  *value = (_regs[reg] >> _pin_shift(pin)) & 1;
}

// Return function and function-specific configuration for a pin.

void
Gpio_x1600_chip::config_pad_get(unsigned pin, unsigned *func, unsigned *value)
{
  unsigned direction, gpio, group0, group1, interrupt, level, trigger;

  config_get(pin, Port_int, &interrupt);

  if (interrupt)
  {
    config_get(pin, Port_trigger, &trigger);
    config_get(pin, Port_level, &level);

    *func = Hw::Gpio_chip::Function_irq;
    *value = (trigger ? (level ? L4_IRQ_F_POS_EDGE : L4_IRQ_F_NEG_EDGE)
                      : (level ? L4_IRQ_F_LEVEL_HIGH : L4_IRQ_F_LEVEL_LOW));
    return;
  }

  config_get(pin, Port_gpio, &gpio);

  if (gpio)
  {
    config_get(pin, Port_dir, &direction);

    *func = Hw::Gpio_chip::Function_gpio;
    *value = direction ? Input : Output;
    return;
  }

  *func = Hw::Gpio_chip::Function_alt;

  config_get(pin, Port_group0, &group0);
  config_get(pin, Port_group1, &group1);

  *value = (group1 << 1) | group0;
}

// Obtain an IRQ abstraction for a pin.

Hw::Gpio_irq_pin *
Gpio_x1600_chip::get_irq(unsigned pin)
{
  if (pin >= _nr_pins)
    throw -L4_EINVAL;

  if (_shadow)
    return new Gpio_x1600_irq_pin(pin, _regs, _shadow_regs, _port_number);
  else
    return new Gpio_x1600_irq_pin(pin, _regs);
}

// Pin function configuration for multiple pins.

void
Gpio_x1600_chip::multi_config_pad(Pin_slice const &mask, unsigned func, unsigned val)
{
  unsigned m = mask.mask;
  for (unsigned pin = mask.offset; pin < _nr_pins; ++pin, m >>= 1)
    if (m & 1)
      config_pad(pin, func, val);
}

// Set up multiple pins with the given mode.

void
Gpio_x1600_chip::multi_setup(Pin_slice const &mask, unsigned mode, unsigned outvalues)
{
  unsigned m = mask.mask;
  for (unsigned pin = mask.offset; pin < _nr_pins; ++pin, m >>= 1, outvalues >>= 1)
    if (m & 1)
      setup(pin, mode, outvalues & 1);
}



// C language interface functions.

void *x1600_gpio_init(l4_addr_t start, l4_addr_t end, unsigned pins,
                      l4_uint32_t pull_ups, l4_uint32_t pull_downs)
{
  return (void *) new Gpio_x1600_chip(start, end, pins, pull_ups, pull_downs);
}

void *x1600_gpio_init_shadow(l4_addr_t start, l4_addr_t end, unsigned pins,
                             l4_uint32_t pull_ups, l4_uint32_t pull_downs,
                             l4_addr_t shadow_start, l4_addr_t shadow_end,
                             uint8_t port_number)
{
  return (void *) new Gpio_x1600_chip(start, end, pins, pull_ups, pull_downs,
                                      shadow_start, shadow_end, port_number);
}

void x1600_gpio_setup(void *gpio, unsigned pin, unsigned mode, int value)
{
  static_cast<Gpio_x1600_chip *>(gpio)->setup(pin, mode, value);
}

void x1600_gpio_config_pull(void *gpio, unsigned pin, unsigned mode)
{
  static_cast<Gpio_x1600_chip *>(gpio)->config_pull(pin, mode);
}

void x1600_gpio_config_pad(void *gpio, unsigned pin, unsigned func, unsigned value)
{
  static_cast<Gpio_x1600_chip *>(gpio)->config_pad(pin, func, value);
}

void x1600_gpio_config_get(void *gpio, unsigned pin, unsigned reg, unsigned *value)
{
  static_cast<Gpio_x1600_chip *>(gpio)->config_get(pin, reg, value);
}

void x1600_gpio_config_pad_get(void *gpio, unsigned pin, unsigned *func, unsigned *value)
{
  static_cast<Gpio_x1600_chip *>(gpio)->config_pad_get(pin, func, value);
}

void x1600_gpio_multi_setup(void *gpio, Pin_slice const *mask, unsigned mode, unsigned outvalues)
{
  static_cast<Gpio_x1600_chip *>(gpio)->multi_setup(*mask, mode, outvalues);
}

void x1600_gpio_multi_config_pad(void *gpio, Pin_slice const *mask, unsigned func, unsigned value)
{
  static_cast<Gpio_x1600_chip *>(gpio)->multi_config_pad(*mask, func, value);
}

void x1600_gpio_multi_set(void *gpio, Pin_slice const *mask, unsigned data)
{
  static_cast<Gpio_x1600_chip *>(gpio)->multi_set(*mask, data);
}

unsigned x1600_gpio_multi_get(void *gpio, unsigned offset)
{
  return static_cast<Gpio_x1600_chip *>(gpio)->multi_get(offset);
}

int x1600_gpio_get(void *gpio, unsigned pin)
{
  return static_cast<Gpio_x1600_chip *>(gpio)->get(pin);
}

void x1600_gpio_set(void *gpio, unsigned pin, int value)
{
  static_cast<Gpio_x1600_chip *>(gpio)->set(pin, value);
}

void *x1600_gpio_get_irq(void *gpio, unsigned pin)
{
  return (void *) static_cast<Gpio_x1600_chip *>(gpio)->get_irq(pin);
}

bool x1600_gpio_irq_set_mode(void *gpio_irq, unsigned mode)
{
  return static_cast<Hw::Gpio_irq_pin *>(gpio_irq)->do_set_mode(mode);
}
